/*******************************************************************************
 * $Id: main.c 206 2008-08-01 21:44:50Z msiegen $
 *
 * Thermostat Controller
 *
 * by Michael Siegenthaler
 * http://www.msigi.net/
 *
 * THIS SOFTWARE IS PUBLIC DOMAIN AND IS PROVIDED WITHOUT ANY WARRANTY
 *
 ******************************************************************************/

#include "omnistat.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

/******************************************************************************/

#define VERSION_STRING "2008-08-01"

#define SYNTAX_ALL_GET "all"
#define SYNTAX_ACTUAL_GET "actual [<C|F|O>]"
#define SYNTAX_OUTSIDE_SET "outside <temperature> [<C|F|O>]"
#define SYNTAX_HEAT_GET "heat [<C|F|O>]"
#define SYNTAX_HEAT_SET "heat <setpoint> [<C|F|O>]"
#define SYNTAX_COOL_GET "cool [<C|F|O>]"
#define SYNTAX_COOL_SET "cool <setpoint> [<C|F|O>]"
#define SYNTAX_MODE_GET "mode"
#define SYNTAX_MODE_SET "mode <off|heat|cool|auto>"
#define SYNTAX_FAN_GET "fan"
#define SYNTAX_FAN_SET "fan <auto|on>"
#define SYNTAX_HOLD_GET "hold"
#define SYNTAX_HOLD_SET "hold <off|on>"
#define SYNTAX_SETCLOCK "setclock"
#define SYNTAX_GROUP_GET "group"
#define SYNTAX_FILTER_GET "filter"
#define SYNTAX_RUNTIME_GET "runtime <current|last>"
#define SYNTAX_OUTPUT_GET "output"
#define SYNTAX_GETREG "getreg <register number>"
#define SYNTAX_SETREG "setreg <register number> <register value>"

#define DEFAULT_UNIT DEG_F

void print_usage() {
  printf("\n  Controller for HAI Omnistat(tm) RC-Series Communicating Thermostats\n\n");
  printf("  Written by Michael Siegenthaler, http://www.msigi.net/\n");
  printf("  Version: %s\n\n", VERSION_STRING);
  printf("  The author of this software is NOT affiliated with HAI in any way.\n");
  printf("  THIS SOFTWARE IS WITHOUT WARRANTY. USE AT YOUR OWN RISK.\n");
  printf("\n  Type 'osc help' for usage.\n\n");
  exit(1);
}

int matches(char* str1, char* str2) {
  return !strcmp(str1, str2);
}

void process_command(char* command, int numargs, char** args) {

  int regnum, value, rv, unit;

  /**************************************************/
  if (matches(command, "help")) {

    printf("\n  Available subcommands:\n");
    printf("    %s\n", SYNTAX_ALL_GET);
    printf("    %s\n", SYNTAX_ACTUAL_GET);
    printf("    %s\n", SYNTAX_OUTSIDE_SET);
    printf("    %s\n", SYNTAX_HEAT_GET);
    printf("    %s\n", SYNTAX_HEAT_SET);
    printf("    %s\n", SYNTAX_COOL_GET);
    printf("    %s\n", SYNTAX_COOL_SET);
    printf("    %s\n", SYNTAX_MODE_GET);
    printf("    %s\n", SYNTAX_MODE_SET);
    printf("    %s\n", SYNTAX_FAN_GET);
    printf("    %s\n", SYNTAX_FAN_SET);
    printf("    %s\n", SYNTAX_HOLD_GET);
    printf("    %s\n", SYNTAX_HOLD_SET);
    printf("    %s\n", SYNTAX_SETCLOCK);
    printf("    %s\n", SYNTAX_GROUP_GET);
    printf("    %s\n", SYNTAX_FILTER_GET);
    printf("    %s\n", SYNTAX_RUNTIME_GET);
    printf("    %s\n", SYNTAX_OUTPUT_GET);
    printf("    %s\n", SYNTAX_GETREG);
    printf("    %s\n", SYNTAX_SETREG);
    printf("\n");

  /**************************************************/
  } else if (matches(command, "getreg")) {

    if ((numargs != 1) || (sscanf(args[0], "%d", &regnum) != 1)) {
      printf("Usage: %s\n", SYNTAX_GETREG);
      exit(1);
    }
    value = readreg(regnum);
    if (value == -1) {
      printf("Could not read register\n");
      exit(1);
    }
    printf("%d\n", value);

  /**************************************************/
  } else if (matches(command, "setreg")) {

    if ((numargs != 2) || (sscanf(args[0], "%d", &regnum) != 1) || (sscanf(args[1], "%d", &value) != 1)) {
      printf("Usage: %s\n", SYNTAX_SETREG);
      exit(1);
    }
    rv = writereg(regnum, value);
    if (rv == -1) {
      printf("Could not set register\n");
      exit(1);
    } else if (rv == 0) {
      printf("Value out of range\n");
      exit(1);
    }

  /**************************************************/
  } else if (matches(command, "setclock")) {

    if (numargs != 0) {
      printf("Usage: %s\n", SYNTAX_SETCLOCK);
      exit(1);
    }

    setclock();

  /**************************************************/
  } else if (matches(command, "mode")) {

    if (numargs == 0) {
      printf("%s\n", mode_text(get_mode()));
    } else if (numargs == 1) {
      value = mode_num(args[0]);
      if (value < 0) {
        printf("Invalid mode\n");
        exit(1);
      }
      rv = set_mode(value);
    } else {
      printf("Usage:\n  Get:  %s\n", SYNTAX_MODE_GET);
      printf("  Set:  %s\n", SYNTAX_MODE_SET);
      exit(1);
    }

  /**************************************************/
  } else if (matches(command, "fan")) {

    if (numargs == 0) {
      printf("%s\n", fan_text(get_fan()));
    } else if (numargs == 1) {
      value = fan_num(args[0]);
      if (value < 0) {
        printf("Invalid fan state\n");
        exit(1);
      }
      rv = set_fan(value);
    } else {
      printf("Usage:\n  Get:  %s\n", SYNTAX_FAN_GET);
      printf("  Set:  %s\n", SYNTAX_FAN_SET);
      exit(1);
    }

  /**************************************************/
  } else if (matches(command, "hold")) {

    if (numargs == 0) {
      printf("%s\n", hold_text(get_hold()));
    } else if (numargs == 1) {
      value = hold_num(args[0]);
      if (value < 0) {
        printf("Invalid hold state\n");
        exit(1);
      }
      rv = set_hold(value);
    } else {
      printf("Usage:\n  Get:  %s\n", SYNTAX_HOLD_GET);
      printf("  Set:  %s\n", SYNTAX_HOLD_SET);
      exit(1);
    }

  /**************************************************/
  } else if (matches(command, "heat")) {

    if (numargs == 0) {
      printf("%d\n", get_heatpoint(DEFAULT_UNIT));
    } else if (numargs == 1) {
      unit = text_to_unit(args[0]);
      if (unit == -1) {
        unit = DEFAULT_UNIT;
        // Could not parse parameter as a unit, so it must be a temperature
        if ( sscanf(args[0], "%d", &value) != 1 )
          goto heat_syntax_err;
        goto heat_set;
      } else {
        printf("%d\n", get_heatpoint(unit));
      }
    } else if (numargs == 2) {
      unit = text_to_unit(args[1]);
      if ((unit == -1) || (sscanf(args[0], "%d", &value) != 1))
        goto heat_syntax_err;
      heat_set:
      rv = set_heatpoint(value, unit);
      if (rv == -1) {
        printf("Could not set temperature\n");
        exit(1);
      } else if (rv == 0) {
        printf("Temperature out of range\n");
        exit(1);
      }
    } else {
      heat_syntax_err:
      printf("Usage:\n  Get:  %s\n", SYNTAX_HEAT_GET);
      printf("  Set:  %s\n", SYNTAX_HEAT_SET);
      exit(1);
    }

  /**************************************************/
  } else if (matches(command, "cool")) {

    if (numargs == 0) {
      printf("%d\n", get_coolpoint(DEFAULT_UNIT));
    } else if (numargs == 1) {
      unit = text_to_unit(args[0]);
      if (unit == -1) {
        unit = DEFAULT_UNIT;
        // Could not parse parameter as a unit, so it must be a temperature
        if ( sscanf(args[0], "%d", &value) != 1 )
          goto cool_syntax_err;
        goto cool_set;
      } else {
        printf("%d\n", get_coolpoint(unit));
      }
    } else if (numargs == 2) {
      unit = text_to_unit(args[1]);
      if ((unit == -1) || (sscanf(args[0], "%d", &value) != 1))
        goto cool_syntax_err;
      cool_set:
      rv = set_coolpoint(value, unit);
      if (rv == -1) {
        printf("Could not set temperature\n");
        exit(1);
      } else if (rv == 0) {
        printf("Temperature out of range\n");
        exit(1);
      }
    } else {
      cool_syntax_err:
      printf("Usage:\n  Get:  %s\n", SYNTAX_COOL_GET);
      printf("  Set:  %s\n", SYNTAX_COOL_SET);
      exit(1);
    }

  /**************************************************/
  } else if (matches(command, "actual")) {

    if (numargs == 0) {
      printf("%d\n", get_actual(DEFAULT_UNIT));
    } else if (numargs == 1) {
      unit = text_to_unit(args[0]);
      if (unit == -1)
        goto actual_syntax_err;
      else
        printf("%d\n", get_actual(unit));
    } else {
      actual_syntax_err:
      printf("Usage:  %s\n", SYNTAX_ACTUAL_GET);
      exit(1);
    }

  /**************************************************/
  } else if (matches(command, "outside")) {

    if (numargs == 1) {
      if ( sscanf(args[0], "%d", &value) != 1 ) {
        goto outside_syntax_err;
      }
      unit = DEFAULT_UNIT;
      goto set_outside;
    } else if (numargs == 2) {
      if ( sscanf(args[0], "%d", &value) != 1 ) {
        goto outside_syntax_err;
      }
      unit = text_to_unit(args[1]);
      if (unit == -1)
        goto outside_syntax_err;
      set_outside:
      rv = set_outside(value, unit);
      if (rv != 1) {
        printf("Cannot set outside temperature\n");
      }
    } else {
      outside_syntax_err:
      printf("Usage:  %s\n", SYNTAX_OUTSIDE_SET);
      exit(1);
    }

  /**************************************************/
  } else if (matches(command, "group")) {

    struct group1_data data;

    if (numargs == 0) {
      unit = DEFAULT_UNIT;
      goto get_group;
    } else if (numargs == 1) {
      unit = text_to_unit(args[0]);
      if (unit == -1)
        goto group_syntax_err;
      get_group:
      rv = poll_group1(unit, &data);
      if (rv == -1) {
        printf("Unable to get data\n");
        exit(1);
      }
      printf("%d %d %s %s %s %d\n", data.coolpoint, data.heatpoint,
        mode_text(data.mode), fan_text(data.fan), hold_text(data.hold), data.actual);
    } else {
      group_syntax_err:
      printf("Usage:  %s\n", SYNTAX_GROUP_GET);
      exit(1);
    }

  /**************************************************/
  } else if (matches(command, "filter")) {

    if (numargs == 0) {
      value = get_filter();
      if (value == -1) {
        printf("Cannot read days remaining on filter\n");
        exit(1);
      }
      printf("%d\n", value);
    } else {
      printf("Usage:  %s\n", SYNTAX_FILTER_GET);
      exit(1);
    }

  /**************************************************/
  } else if (matches(command, "runtime")) {

    if (numargs == 1) {
      if (matches(args[0], "current")) {
        value = get_current_runtime();
      } else if (matches(args[0], "last")) {
        value = get_last_runtime();
      } else {
        goto runtime_syntax_err;
      }
      if (value == -1) {
        printf("Cannot read runtime\n");
        exit(1);
      }
      printf("%d\n", value);
    } else {
      runtime_syntax_err:
      printf("Usage:  %s\n", SYNTAX_RUNTIME_GET);
      exit(1);
    }

  /**************************************************/
  } else if (matches(command, "output")) {

    if (numargs == 0) {
      value = get_output_status();
      if (value == -1) {
        printf("Cannot read output status\n");
        exit(1);
      }
      printf("%d\n", value);
    } else {
      printf("Usage:  %s\n", SYNTAX_OUTPUT_GET);
      exit(1);
    }

  /**************************************************/
  } else if (matches(command, "all")) {

    struct group1_data data;
    int output_status, filter_status, current_runtime, last_runtime;

    if (numargs == 0) {
      unit = DEFAULT_UNIT;
      goto get_all;
    } else if (numargs == 1) {
      unit = text_to_unit(args[0]);
      if (unit == -1)
        goto all_syntax_err;
      get_all:
      rv = poll_group1(unit, &data);
      if (rv == -1) {
        printf("Unable to get group data\n");
        exit(1);
      }
      printf("\n");
      printf("  Actual temperature:  %d %s\n", data.actual, unit_to_text(unit));
      if (data.mode == MODE_HEAT)
        printf("            Setpoint:  %d %s\n", data.heatpoint, unit_to_text(unit));
      else if (data.mode == MODE_COOL)
        printf("            Setpoint:  %d %s\n", data.coolpoint, unit_to_text(unit));
      printf("      Operating mode:  %s\n", mode_text(data.mode));
      printf("           Hold mode:  %s\n", hold_text(data.hold));
      printf("            Fan mode:  %s\n", fan_text(data.fan));

      output_status = get_output_status();
      if (output_status == -1) {
        printf("Cannot read output status\n");
        exit(1);
      }
      printf("              Output:  %s\n", (output_status & 0x4) ? "on" : "off");

      filter_status = get_filter();
      if (filter_status == -1) {
        printf("Cannot read days remaining on filter\n");
        exit(1);
      }
      printf("              Filter:  %d days remaining\n", filter_status);

      current_runtime = get_current_runtime();
      if (current_runtime == -1) {
        printf("Cannot get current runtime\n");
        exit(1);
      }
      printf(" This week's runtime:  %d hours\n", current_runtime);

      last_runtime = get_last_runtime();
      if (last_runtime == -1) {
        printf("Cannot get last runtime\n");
        exit(1);
      }
      printf(" Last week's runtime:  %d hours\n", last_runtime);

      printf("\n");
    } else {
      all_syntax_err:
      printf("Usage:  %s\n", SYNTAX_ALL_GET);
      exit(1);
    }

  /**************************************************/
  } else {

    printf("Unrecognized command. Type 'osc help' for usage.\n");
    exit(1);

  }

}

int main(int argc, char ** argv) {


  if (argc < 2) {
    print_usage();
  }

  process_command(argv[1], argc-2, argv+2);

  return 0;

}
