/*******************************************************************************
 * $Id: omnistat.c 206 2008-08-01 21:44:50Z msiegen $
 *
 * Thermostat API
 *
 * by Michael Siegenthaler
 * http://www.msigi.net/
 *
 * THIS SOFTWARE IS PUBLIC DOMAIN AND IS PROVIDED WITHOUT ANY WARRANTY
 *
 ******************************************************************************/

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>

#include <time.h>

#include <stdio.h>
#include <unistd.h>
#include <string.h>

#include <limits.h>

#include <termios.h>

#include "config.h" // must come before other includes
#include "network.h"
#include "serial.h"

#include "omnistat.h"

/******************************************************************************/

#define BUFLEN 100

int initialized = 0;

// Send a message to the thermostat and receive a reply
// Return input length on success, -1 on failure
int query(char* outbuf, int outlen, char* inbuf, int in_maxlen) {

  int i, rv;

  if (!initialized) {
    if ( initialize_port() == -1 ) return -1;
    initialized = 1;
  }

  for (i=0; i<=MAX_RETRIES; i++) {

    rv = send_command(outbuf, outlen);
    if (rv == -1) continue;

    rv = receive_reply(inbuf, in_maxlen);
    if (rv == -1) continue;

    return rv;
  }

  return -1;
}


// Read a register
// Return the register value on success, or -1 on failure
int readreg(int regnum) {

  int i, inlen;
  char outbuf[BUFLEN];
  char inbuf[BUFLEN];

  outbuf[0] = TSTAT_ADDR;
  outbuf[1] = 0x20; // data length = 2, message type = 0
  outbuf[2] = regnum & 0xff; // register number
  outbuf[3] = 1; // number of registers
  outbuf[4] = 0;
  for (i=0; i<4; i++)
    outbuf[4] += (outbuf[i] & 0xff); // checksum

  inlen = query(outbuf, 5, inbuf, BUFLEN);

  if (inlen == -1) {
    return -1;
  }

  return inbuf[3] & 0xff;

}

// Write a register
// Return 1 if register has been set, 0 if negative ack, -1 if error
int writereg(int regnum, int value) {

  int i, inlen;
  char outbuf[BUFLEN];
  char inbuf[BUFLEN];

  outbuf[0] = TSTAT_ADDR;
  outbuf[1] = 0x21; // data length = 2, message type = 1
  outbuf[2] = regnum & 0xff; // register number
  outbuf[3] = value & 0xff; // new value
  outbuf[4] = 0;
  for (i=0; i<4; i++)
    outbuf[4] += (outbuf[i] & 0xff); // checksum

  inlen = query(outbuf, 5, inbuf, BUFLEN);

  if (inlen == -1)
    return -1; // could not communicate with thermostat

  if (inlen < 3)
    return -1; // reply message too short

  if (inbuf[1] == 0x00)
    return 1; // positive ack

  if (inbuf[1] == 0x01)
    return 0; // negative ack

  return -1; // some other problem

}

// Poll for group 1 data
// Return the register value on success, or -1 on failure
int poll_group1(int unit, struct group1_data* data) {

  int inlen;
  char outbuf[BUFLEN];
  char inbuf[BUFLEN];

  outbuf[0] = TSTAT_ADDR;
  outbuf[1] = 0x02; // data length = 0, message type = 2
  outbuf[2] = (outbuf[0] & 0xff) + (outbuf[1] & 0xff); // checksum

  inlen = query(outbuf, 3, inbuf, BUFLEN);

  if (inlen == -1) {
    return -1;
  }

  if (inlen != 9) {
    return -1;
  }

  data->unit = unit;
  data->mode = inbuf[4] & 0xff;
  data->fan = inbuf[5] & 0xff;
  data->hold = inbuf[6] & 0xff;

  if (unit == DEG_F) {
    data->coolpoint = o_to_f(inbuf[2] & 0xff);
    data->heatpoint = o_to_f(inbuf[3] & 0xff);
    data->actual = o_to_f(inbuf[7] & 0xff);
  } else if (unit == DEG_C) {
    data->coolpoint = o_to_c(inbuf[2] & 0xff);
    data->heatpoint = o_to_c(inbuf[3] & 0xff);
    data->actual = o_to_c(inbuf[7] & 0xff);
  } else if (unit == DEG_O) {
    data->coolpoint = inbuf[2] & 0xff;
    data->heatpoint = inbuf[3] & 0xff;
    data->actual = inbuf[7] & 0xff;
  } else {
    return -1;
  }

  return 0;

}

// Set clock
// Return 0 on success, -1 on failure
int setclock() {

  int i, inlen;
  char outbuf[BUFLEN];
  char inbuf[BUFLEN];
  struct tm * now;
  int day_of_week;
  time_t currtime;

  currtime = time(0);
  now = localtime(&currtime);
  day_of_week = (now->tm_wday + 6) % 7;

  //printf("Setting clock: %02d:%02d:%02d, day_of_week=%d\n", now->tm_hour, now->tm_min, now->tm_sec, day_of_week);

  outbuf[0] = TSTAT_ADDR;
  outbuf[1] = 0x41; // data length = 4, message type = 1
  outbuf[2] = 0x41; // start at this register
  outbuf[3] = now->tm_sec; // BUG: this will be 60 during a leap-second, but that's rare
  outbuf[4] = now->tm_min;
  outbuf[5] = now->tm_hour;
  outbuf[6] = 0;
  for (i=0; i<6; i++)
    outbuf[6] += (outbuf[i] & 0xff); // checksum

  inlen = query(outbuf, 7, inbuf, BUFLEN);

  if (inlen == -1)
    return -1; // could not communicate with thermostat

  if (inlen < 3)
    return -1; // reply message too short

  if (inbuf[1] == 0x01)
    return 0; // negative ack

  // Also set the day of the week
  return writereg(0x3a, day_of_week);
  // BUG: race condition if the clock already rolled over to the next day

}

// Set mode (MODE_OFF, MODE_HEAT, MODE_COOL, MODE_AUTO)
// Return 1 on success, 0 if out of range, -1 on failure
int set_mode(int mode) {
  if ((mode < MODE_OFF) || (mode > MODE_AUTO))
    return 0;
  return writereg(0x3d, mode);
}

// Get mode (MODE_OFF, MODE_HEAT, MODE_COOL, MODE_AUTO)
// Returns mode on success, -1 on failure
int get_mode() {
  return readreg(0x3d);
}

// Convert mode number to text
char* mode_text(int num) {
  if (num == MODE_OFF)
    return "off";
  else if (num == MODE_HEAT)
    return "heat";
  else if (num == MODE_COOL)
    return "cool";
  else if (num == MODE_AUTO)
    return "auto";
  else
    return "invalid";
}

// Convert mode text to number
int mode_num(char* text) {
  if (!strncmp(text, "off", 3))
    return MODE_OFF;
  else if (!strncmp(text, "heat", 4))
    return MODE_HEAT;
  else if (!strncmp(text, "cool", 4))
    return MODE_COOL;
  else if (!strncmp(text, "auto", 4))
    return MODE_AUTO;
  else
    return -1;
}

// Set fan (FAN_AUTO, FAN_ON)
// Return 1 on success, 0 if out of range, -1 on failure
int set_fan(int fan) {
  if ((fan < FAN_AUTO) || (fan > FAN_ON))
    return 0;
  return writereg(0x3e, fan);
}

// Get fan status (FAN_AUTO, FAN_ON)
// Returns status on success, -1 on failure
int get_fan() {
  return readreg(0x3e);
}

// Convert fan status number to text
char* fan_text(int num) {
  if (num == FAN_AUTO)
    return "auto";
  else if (num == FAN_ON)
    return "on";
  else
    return "invalid";
}

// Convert fan status text to number
int fan_num(char* text) {
  if (!strncmp(text, "auto", 4))
    return FAN_AUTO;
  else if (!strncmp(text, "on", 2))
    return FAN_ON;
  else
    return -1;
}

// Set hold (HOLD_OFF, HOLD_ON)
// Return 1 on success, 0 if out of range, -1 on failure
int set_hold(int hold) {
  if ((hold != HOLD_OFF) && (hold != HOLD_ON))
    return 0;
  return writereg(0x3f, hold);
}

// Get hold status (HOLD_OFF, HOLD_ON)
// Returns status on success, -1 on failure
int get_hold() {
  return readreg(0x3f);
}

// Convert hold status number to text
char* hold_text(int num) {
  if (num == HOLD_OFF)
    return "off";
  else if (num == HOLD_ON)
    return "on";
  else
    return "invalid";
}

// Convert hold status text to number
int hold_num(char* text) {
  if (!strncmp(text, "off", 3))
    return HOLD_OFF;
  else if (!strncmp(text, "on", 2))
    return HOLD_ON;
  else
    return -1;
}

// Days remaining until filter reminder
// Returns value on success, -1 on failure
int get_filter() {
  return readreg(0x0f);
}

// System runtime, current week - hours
// Returns value on success, -1 on failure
int get_current_runtime() {
  return readreg(0x10);
}

// System runtime, last week - hours
// Returns value on success, -1 on failure
int get_last_runtime() {
  return readreg(0x11);
}

// Output status
// Returns value on success, -1 on failure
int get_output_status() {
  return readreg(0x48);
}

// Set the heat setpoint. Unit may be DEG_F, DEG_C, or DEG_O
// Return 1 on success, 0 if out of range, -1 on failure
int set_heatpoint(int value, int unit) {
  return set_some_temp(0x3c, value, unit);
}

// Get the heat setpoint. Unit may be DEG_F, DEG_C, or DEG_O
// Return value on success, 255 on failure
int get_heatpoint(int unit) {
  return get_some_temp(0x3c, unit);
}

// Set the cool setpoint. Unit may be DEG_F, DEG_C, or DEG_O
// Return 1 on success, 0 if out of range, -1 on failure
int set_coolpoint(int value, int unit) {
  return set_some_temp(0x3b, value, unit);
}

// Get the cool setpoint. Unit may be DEG_F, DEG_C, or DEG_O
// Return value on success, 255 on failure
int get_coolpoint(int unit) {
  return get_some_temp(0x3b, unit);
}

// Set the outside temperature. Unit may be DEG_F, DEG_C, or DEG_O
// Return 1 on success, 0 if out of range, -1 on failure
int set_outside(int value, int unit) {
  return set_some_temp(0x44, value, unit);
}

// Get actual temperature
// Return temperature on success, 255 on failure
int get_actual(int unit) {
  return get_some_temp(0x40, unit);
}

// Set a temperature
// Return 1 on success, 0 if out of range, -1 on error
int set_some_temp(int regnum, int value, int unit) {
  if (unit == DEG_F)
    value = f_to_o(value);
  else if (unit == DEG_C)
    value = c_to_o(value);
  else if (unit != DEG_O)
    return 0;
  return writereg(regnum, value);
}

// Get a temperature
// Return 255 on error
int get_some_temp(int regnum, int unit) {
  int val;
  val = readreg(regnum);
  if (val == -1) return 255;
  if (unit == DEG_F)
    return o_to_f(val);
  else if (unit == DEG_C)
    return o_to_c(val);
  else if (unit != DEG_O)
    return 255;
  return val;
}

// Convert Omnistat(tm) native temperature unit to degrees Celsius
int o_to_c(int dego) {
  double degc;
  degc = (((double)dego)/2.0)-40.0;
  return (int)(degc+0.5);
  // BUG: rounding error when degc is negative, but this is an indoor thermostat, so who cares??
}

// Convert degrees Celsius to Omnistat(tm) native temperature unit
int c_to_o(int degc) {
  return (degc+40)*2;
}

// Convert Omnistat(tm) native temperature unit to degrees Fahrenheit
int o_to_f(int dego) {
  double degf;
  degf = (9.0/10.0)*((double)(dego-80)) + 32.0;
  return (int)(degf+0.5);
  // BUG: rounding error when degf is negative, but this is an indoor thermostat, so who cares??
}

// Convert degrees Fahrenheit to Omnistat(tm) native temperature unit
int f_to_o(int degf) {
  double dego;
  dego = (10.0/9.0)*((double)(degf-32)) + 80.0;
  return (int)(dego+0.5);
}

// Convert text string into a unit
int text_to_unit(char* text) {

  switch (text[0]) {
    case 'C':
    case 'c':
      return DEG_C;
    case 'F':
    case 'f':
      return DEG_F;
    case 'O':
    case 'o':
      return DEG_O;
    default:
      return -1;
  }

}

// Convert a unit into a text string
char* unit_to_text(int unit) {

  switch (unit) {
    case DEG_C:
      return "C";
    case DEG_F:
      return "F";
    case DEG_O:
      return "O";
    default:
      return "unknown unit";
  }

}
