/*******************************************************************************
 * $Id: serial.c 206 2008-08-01 21:44:50Z msiegen $
 *
 * Thermostat Serial Interface
 *
 * by Michael Siegenthaler
 * http://www.msigi.net/
 *
 * THIS SOFTWARE IS PUBLIC DOMAIN AND IS PROVIDED WITHOUT ANY WARRANTY
 *
 ******************************************************************************/

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/file.h>
#include <fcntl.h>

#include <stdio.h>
#include <unistd.h>
#include <string.h>

#include <limits.h>

#include <termios.h>

#include "config.h" // must come before other includes
#include "serial.h"

/******************************************************************************/

int serial_fd = -1;
int lock_fd = -1;

// Initialize the serial port
// Return 0 on success, -1 on failure
int initialize_port_ser() {

  struct termios newtio;

  // Open serial port
  serial_fd = open(SERIALDEVICE, O_RDWR | O_NOCTTY);
  if (serial_fd < 0) {
    return -1;
  }

  // Open the lockfile
  lock_fd = open(LOCKFILE, O_CREAT | O_RDWR | O_NOCTTY, 0644);
  if (lock_fd < 0) {
    lock_fd = open(LOCKFILE, O_RDONLY | O_NOCTTY);
    if (lock_fd < 0) {
      printf("Cannot open lockfile\n");
      return -1;
    }
  }

  // Lock the serial port
  if ( flock(lock_fd, LOCK_EX) == -1 ) {
    printf("Cannot get lock\n");
    return -1;
  }
  // Possible BUG: We never unlock the port, or close the file descriptor, because
  // the assumption is that the program is short-lived and will not block on anything
  // other than the serial port itself. If this behavior is later changed, it may
  // become necessary to unlock the port.

  // Specify new port settings
  memset(&newtio, 0, sizeof(newtio));
  newtio.c_cflag = CS8 | CLOCAL | CREAD;
#ifdef FLOWCONTROL_RTSCTS
  newtio.c_cflag |= CRTSCTS;
#endif
  newtio.c_iflag = IGNPAR; // If using a terminal, ICRNL may be included here as well
#ifdef FLOWCONTROL_XONXOFF
  newtio.c_iflag |= IXON | IXOFF;
#endif
  newtio.c_oflag = 0;
  newtio.c_lflag = 0; // non-canonical input
  cfsetspeed(&newtio, BAUDRATE);
  newtio.c_cc[VINTR]    = 0;     /* Ctrl-c */ 
  newtio.c_cc[VQUIT]    = 0;     /* Ctrl-\ */
  newtio.c_cc[VERASE]   = 0;     /* del */
  newtio.c_cc[VKILL]    = 0;     /* @ */
  newtio.c_cc[VEOF]     = 4;     /* Ctrl-d */
  newtio.c_cc[VTIME]    = 0;     /* inter-character timer unused */
  newtio.c_cc[VMIN]     = 1;     /* blocking read until 1 character arrives */
  //newtio.c_cc[VSWTC]    = 0;     /* '\0' (this field does not exist on FreeBSD or Mac OS X) */
  newtio.c_cc[VSTART]   = 0;     /* Ctrl-q */ 
  newtio.c_cc[VSTOP]    = 0;     /* Ctrl-s */
  newtio.c_cc[VSUSP]    = 0;     /* Ctrl-z */
  newtio.c_cc[VEOL]     = 0;     /* '\0' */
  newtio.c_cc[VREPRINT] = 0;     /* Ctrl-r */
  newtio.c_cc[VDISCARD] = 0;     /* Ctrl-u */
  newtio.c_cc[VWERASE]  = 0;     /* Ctrl-w */
  newtio.c_cc[VLNEXT]   = 0;     /* Ctrl-v */
  newtio.c_cc[VEOL2]    = 0;     /* '\0' */

  // Activate new port settings
  tcflush(serial_fd, TCIOFLUSH);
  tcsetattr(serial_fd, TCSANOW, &newtio);

  return 0;
}

// Send a command to the thermostat
// Return 0 on success, -1 on failure
int send_command_ser(char* buf, int length) {

  int bytes_written;
  int bytes_already_written = 0;

  // Make sure port has been opened
  if (serial_fd < 0) return -1;

  // Flush the I/O buffers, in case a previous command still has reply data waiting
  tcflush(serial_fd, TCIOFLUSH);

  // Write the buffer out to the serial port in a loop
  while (bytes_already_written < length) {
    // Write to port
    //printf("Writing %02x %02x %02x %02x %02x\n", buf[0] & 0xff, buf[1] & 0xff, buf[2] & 0xff, buf[3] & 0xff, buf[4] & 0xff);
    bytes_written = write(serial_fd, buf+bytes_already_written, length-bytes_already_written);
    if (bytes_written < 0 )
      return -1;
    bytes_already_written += bytes_written;
  }

  return 0;
}

// Receive a reply from the thermostat
// Return length of received reply, or -1 on error
int receive_reply_ser(char* buf, int maxlen) {

  int i;
  int bytes_read;
  int bytes_already_read;
  struct timeval tv;
  fd_set readfds;
  int msg_length = 0; // refers to data section only
  int cksum;

  // Make sure port has been opened
  if (serial_fd < 0) return -1;

  // Make sure the buffer is long enough for the shortest possible message
  if (maxlen < 3) return -1;

  bytes_already_read = 0;

  // set timeout for read from serial port
  tv.tv_sec = 1; // set read timeout to 1.5 seconds
  tv.tv_usec = 500000;

  while (1) {

    FD_ZERO(&readfds);
    FD_SET(serial_fd, &readfds);

    // Check if data is available on serial port
    if ( select(serial_fd+1, &readfds, NULL, NULL, &tv) <= 0 ) {
      // Timed out, or error
      return -1;
    }

    if (FD_ISSET(serial_fd, &readfds)) { // we have input

      // Read from serial port
      bytes_read = read(serial_fd, buf+bytes_already_read, maxlen-bytes_already_read);

      //printf("read %d bytes from serial port: ", bytes_read);
      //for (i=0; i<bytes_read; i++) {
      //  printf("0x%02x ", *(unsigned char*)(buf+bytes_already_read+i));
      //}
      //printf("\n");

      if (bytes_read < 0)
        return -1;

      bytes_already_read += bytes_read;

      if (bytes_already_read >= 2) {
        // Update the message length
        msg_length = (buf[1] >> 4) & 0x0f;
      }

      if (bytes_already_read >= msg_length+3) {
        // Complete message received...verify the checksum
        cksum = 0;
        for (i=0; i<msg_length+2; i++)
          cksum += (buf[i] & 0xff);
        if ((cksum & 0xff) != (buf[msg_length+2] & 0xff)) {
          //printf("Checksum does not match: should be %02x\n", cksum & 0xff);
          return -1; // checksum mismatch!
        } else {
          // Good!
          return bytes_already_read;
        }
      }
    }
  }

  return -1;
}
